/* -*- c++ -*-
 * Copyright (c) 2012-2026 by the GalSim developers team on GitHub
 * https://github.com/GalSim-developers
 *
 * This file is part of GalSim: The modular galaxy image simulation toolkit.
 * https://github.com/GalSim-developers/GalSim
 *
 * GalSim is free software: redistribution and use in source and binary forms,
 * with or without modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions, and the disclaimer given in the accompanying LICENSE
 *    file.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions, and the disclaimer given in the documentation
 *    and/or other materials provided with the distribution.
 */

#include "PyBind11Helper.h"
#include "math/Bessel.h"
#include "math/Gamma.h"
#include "math/Sinc.h"

namespace galsim {
namespace math {

    void pyExportBessel(py::module& _galsim)
    {
        _galsim.def("j0_root", &getBesselRoot0);
        _galsim.def("jv_root", &getBesselRoot);
        _galsim.def("j0", &j0);
        _galsim.def("j1", &j1);
        _galsim.def("jv", &cyl_bessel_j);
        _galsim.def("yv", &cyl_bessel_y);
        _galsim.def("iv", &cyl_bessel_i);
        _galsim.def("kv", &cyl_bessel_k);

        // Include a few other (non-Bessel) math items of similar utility.
        _galsim.def("gammainc", &gamma_p);
        _galsim.def("sinc", &sinc);
        _galsim.def("si", &Si);
        _galsim.def("ci", &Ci);
    }

} // namespace math
} // namespace galsim

